@@ -64,7 +64,6 @@ use std::{
 
 use crc16::*;
 use futures::{
-    channel::{mpsc, oneshot},
     future::{self, BoxFuture},
     prelude::*,
     ready, stream,
@@ -77,6 +76,7 @@ use redis::{
     aio::ConnectionLike, Cmd, ConnectionAddr, ConnectionInfo, ErrorKind, IntoConnectionInfo,
     RedisError, RedisFuture, RedisResult, Value,
 };
+use tokio::sync::{mpsc, oneshot};
 
 const SLOT_SIZE: usize = 16384;
 const DEFAULT_RETRIES: u32 = 16;
@@ -99,7 +99,7 @@ impl Client {
 
         for info in initial_nodes {
             let info = info.into_connection_info()?;
-            if let ConnectionAddr::Unix(_) = *info.addr {
+            if let ConnectionAddr::Unix(_) = info.addr {
                 return Err(RedisError::from((ErrorKind::InvalidClientConfig,
                                              "This library cannot use unix socket because Redis's cluster command returns only cluster's IP and port.")));
             }
@@ -131,7 +131,7 @@ impl Client {
     #[doc(hidden)]
     pub async fn get_generic_connection<C>(&self) -> RedisResult<Connection<C>>
     where
-        C: ConnectionLike + Connect + Clone + Send + Unpin + 'static,
+        C: ConnectionLike + Connect + Clone + Send + Sync + Unpin + 'static,
     {
         Connection::new(&self.initial_nodes, self.retries).await
     }
@@ -143,24 +143,33 @@ pub struct Connection<C = redis::aio::MultiplexedConnection>(mpsc::Sender<Messag
 
 impl<C> Connection<C>
 where
-    C: ConnectionLike + Connect + Clone + Send + Unpin + 'static,
+    C: ConnectionLike + Connect + Clone + Send + Sync + Unpin + 'static,
 {
     async fn new(
         initial_nodes: &[ConnectionInfo],
         retries: Option<u32>,
     ) -> RedisResult<Connection<C>> {
         Pipeline::new(initial_nodes, retries).await.map(|pipeline| {
-            let (tx, rx) = mpsc::channel::<Message<_>>(100);
-            tokio::spawn(rx.map(Ok).forward(pipeline).map(|_| ()));
+            let (tx, mut rx) = mpsc::channel::<Message<_>>(100);
+
+            tokio::spawn(async move {
+                let _ = stream::poll_fn(move |cx| rx.poll_recv(cx))
+                    .map(Ok)
+                    .forward(pipeline)
+                    .await;
+            });
+
             Connection(tx)
         })
     }
 }
 
 type SlotMap = BTreeMap<u16, String>;
+type ConnectionFuture<C> = future::Shared<BoxFuture<'static, C>>;
+type ConnectionMap<C> = HashMap<String, ConnectionFuture<C>>;
 
 struct Pipeline<C> {
-    connections: HashMap<String, C>,
+    connections: ConnectionMap<C>,
     slots: SlotMap,
     state: ConnectionState<C>,
     in_flight_requests: stream::FuturesUnordered<
@@ -256,7 +265,7 @@ struct Message<C> {
 }
 
 type RecoverFuture<C> =
-    BoxFuture<'static, Result<(SlotMap, HashMap<String, C>), (RedisError, HashMap<String, C>)>>;
+    BoxFuture<'static, Result<(SlotMap, ConnectionMap<C>), (RedisError, ConnectionMap<C>)>>;
 
 enum ConnectionState<C> {
     PollComplete,
@@ -422,7 +431,7 @@ where
 
 impl<C> Pipeline<C>
 where
-    C: ConnectionLike + Connect + Clone + Send + 'static,
+    C: ConnectionLike + Connect + Clone + Send + Sync + 'static,
 {
     async fn new(initial_nodes: &[ConnectionInfo], retries: Option<u32>) -> RedisResult<Self> {
         let connections = Self::create_initial_connections(initial_nodes).await?;
@@ -443,11 +452,11 @@ where
 
     async fn create_initial_connections(
         initial_nodes: &[ConnectionInfo],
-    ) -> RedisResult<HashMap<String, C>> {
+    ) -> RedisResult<ConnectionMap<C>> {
         let connections = stream::iter(initial_nodes.iter().cloned())
             .map(|info| async move {
-                let addr = match *info.addr {
-                    ConnectionAddr::Tcp(ref host, port) => match &info.passwd {
+                let addr = match info.addr {
+                    ConnectionAddr::Tcp(ref host, port) => match &info.redis.password {
                         Some(pw) => format!("redis://:{}@{}:{}", pw, host, port),
                         None => format!("redis://{}:{}", host, port),
                     },
@@ -456,14 +465,14 @@ where
 
                 let result = connect_and_check(info).await;
                 match result {
-                    Ok(conn) => Some((addr, conn)),
+                    Ok(conn) => Some((addr, async { conn }.boxed().shared())),
                     Err(_) => None,
                 }
             })
             .buffer_unordered(initial_nodes.len())
             .fold(
                 HashMap::with_capacity(initial_nodes.len()),
-                |mut connections: HashMap<String, C>, conn: Option<(String, C)>| async move {
+                |mut connections: ConnectionMap<C>, conn| async move {
                     connections.extend(conn);
                     connections
                 },
@@ -481,14 +490,15 @@ where
     // Query a node to discover slot-> master mappings.
     fn refresh_slots(
         &mut self,
-    ) -> impl Future<Output = Result<(SlotMap, HashMap<String, C>), (RedisError, HashMap<String, C>)>>
+    ) -> impl Future<Output = Result<(SlotMap, ConnectionMap<C>), (RedisError, ConnectionMap<C>)>>
     {
         let mut connections = mem::replace(&mut self.connections, Default::default());
 
         async move {
             let mut result = Ok(SlotMap::new());
             for (addr, conn) in connections.iter_mut() {
-                match get_slots(addr, &mut *conn)
+                let mut conn = conn.clone().await;
+                match get_slots(addr, &mut conn)
                     .await
                     .and_then(|v| Self::build_slot_map(v))
                 {
@@ -512,7 +522,8 @@ where
                     (connections, new_connections),
                     move |(mut connections, mut new_connections), addr| async move {
                         if !new_connections.contains_key(addr) {
-                            let new_connection = if let Some(mut conn) = connections.remove(addr) {
+                            let new_connection = if let Some(conn) = connections.remove(addr) {
+                                let mut conn = conn.await;
                                 match check_connection(&mut conn).await {
                                     Ok(_) => Some((addr.to_string(), conn)),
                                     Err(_) => match connect_and_check(addr.as_ref()).await {
@@ -526,7 +537,10 @@ where
                                     Err(_) => None,
                                 }
                             };
-                            new_connections.extend(new_connection);
+                            if let Some((addr, new_connection)) = new_connection {
+                                new_connections
+                                    .insert(addr, async { new_connection }.boxed().shared());
+                            }
                         }
                         (connections, new_connections)
                     },
@@ -567,50 +581,48 @@ where
         Ok(slot_map)
     }
 
-    fn get_connection(&self, slot: u16) -> impl Future<Output = (String, C)> + 'static {
+    fn get_connection(&mut self, slot: u16) -> (String, ConnectionFuture<C>) {
         if let Some((_, addr)) = self.slots.range(&slot..).next() {
-            if self.connections.contains_key(addr) {
-                return future::Either::Left(future::ready((
-                    addr.clone(),
-                    self.connections.get(addr).unwrap().clone(),
-                )));
+            if let Some(conn) = self.connections.get(addr) {
+                return (addr.clone(), conn.clone());
             }
 
             // Create new connection.
             //
-            let random_conn = get_random_connection(&self.connections, None); // TODO Only do this lookup if the first check fails
-            let addr = addr.clone();
-            future::Either::Right(async move {
-                let result = connect_and_check(addr.as_ref()).await;
-                result
-                    .map(|conn| (addr, conn))
-                    .unwrap_or_else(|_| random_conn)
-            })
+            let (_, random_conn) = get_random_connection(&self.connections, None); // TODO Only do this lookup if the first check fails
+            let connection_future = {
+                let addr = addr.clone();
+                async move {
+                    match connect_and_check(addr.as_ref()).await {
+                        Ok(conn) => conn,
+                        Err(_) => random_conn.await,
+                    }
+                }
+            }
+            .boxed()
+            .shared();
+            self.connections
+                .insert(addr.clone(), connection_future.clone());
+            (addr.clone(), connection_future)
         } else {
             // Return a random connection
-            future::Either::Left(future::ready(get_random_connection(
-                &self.connections,
-                None,
-            )))
+            get_random_connection(&self.connections, None)
         }
     }
 
     fn try_request(
-        &self,
+        &mut self,
         info: &RequestInfo<C>,
     ) -> impl Future<Output = (String, RedisResult<Response>)> {
         // TODO remove clone by changing the ConnectionLike trait
         let cmd = info.cmd.clone();
-        let f = if info.excludes.len() > 0 || info.slot.is_none() {
-            future::Either::Left(future::ready(get_random_connection(
-                &self.connections,
-                Some(&info.excludes),
-            )))
+        let (addr, conn) = if info.excludes.len() > 0 || info.slot.is_none() {
+            get_random_connection(&self.connections, Some(&info.excludes))
         } else {
-            future::Either::Right(self.get_connection(info.slot.unwrap()))
+            self.get_connection(info.slot.unwrap())
         };
         async move {
-            let (addr, conn) = f.await;
+            let conn = conn.await;
             let result = cmd.exec(conn).await;
             (addr, result)
         }
@@ -648,6 +660,13 @@ where
         if !self.pending_requests.is_empty() {
             let mut pending_requests = mem::take(&mut self.pending_requests);
             for request in pending_requests.drain(..) {
+                // Drop the request if noone is waiting for a response to free up resources for
+                // requests callers care about (load shedding). It will be ambigous whether the
+                // request actually goes through regardless.
+                if request.sender.is_closed() {
+                    continue;
+                }
+
                 let future = self.try_request(&request.info);
                 self.in_flight_requests.push(Box::pin(Request {
                     max_retries: self.retries,
@@ -718,7 +737,7 @@ where
 
 impl<C> Sink<Message<C>> for Pipeline<C>
 where
-    C: ConnectionLike + Connect + Clone + Send + Unpin + 'static,
+    C: ConnectionLike + Connect + Clone + Send + Sync + Unpin + 'static,
 {
     type Error = ();
 
@@ -921,9 +940,6 @@ impl Clone for Client {
     }
 }
 
-trait ImplRedisFuture<T>: Future<Output = RedisResult<T>> {}
-impl<T, F> ImplRedisFuture<T> for F where F: Future<Output = RedisResult<T>> {}
-
 pub trait Connect: Sized {
     fn connect<'a, T>(info: T) -> RedisFuture<'a, Self>
     where
@@ -944,15 +960,14 @@ impl Connect for redis::aio::MultiplexedConnection {
     }
 }
 
-fn connect_and_check<'a, T, C>(info: T) -> impl ImplRedisFuture<C> + 'a
+async fn connect_and_check<T, C>(info: T) -> RedisResult<C>
 where
-    T: IntoConnectionInfo + Send + 'a,
+    T: IntoConnectionInfo + Send,
     C: ConnectionLike + Connect + Send + 'static,
 {
-    C::connect(info).and_then(|mut conn| async move {
-        check_connection(&mut conn).await?;
-        Ok(conn)
-    })
+    let mut conn = C::connect(info).await?;
+    check_connection(&mut conn).await?;
+    Ok(conn)
 }
 
 async fn check_connection<C>(conn: &mut C) -> RedisResult<()>
@@ -966,9 +981,9 @@ where
 }
 
 fn get_random_connection<'a, C>(
-    connections: &'a HashMap<String, C>,
+    connections: &'a ConnectionMap<C>,
     excludes: Option<&'a HashSet<String>>,
-) -> (String, C)
+) -> (String, ConnectionFuture<C>)
 where
     C: Clone,
 {
@@ -1121,9 +1136,7 @@ where
 }
 
 fn get_password(addr: &str) -> Option<String> {
-    redis::parse_redis_url(addr)
-        .ok()
-        .and_then(|url| url.password().map(|s| s.into()))
+    redis::parse_redis_url(addr).and_then(|url| url.password().map(|s| s.into()))
 }
 
 #[cfg(test)]
