@@ -33,7 +33,7 @@ impl Connect for MockConnection {
     {
         let info = info.into_connection_info().unwrap();
 
-        let (name, port) = match &*info.addr {
+        let (name, port) = match &info.addr {
             redis::ConnectionAddr::Tcp(addr, port) => (addr, *port),
             _ => unreachable!(),
         };
